from ..socket import ServiceConsumer

'''
'''
class InfoClient:
    def __init__(self, server_address):
        self.service = ServiceConsumer(
            'trodes.info', server_address=server_address)

    def __request(self, item):
        data = { 'request': item }
        return self.service.request(data)

    def request_time(self):
        return self.__request('time')[2]['time']

    def request_timerate(self):
        return self.__request('timerate')[2]['timerate']

    def request_config(self):
        return self.__request('config')


import threading
from ..socket import ServiceConsumer

'''
Use this class to open and close files from Python.

Requires a server_address to connect to the server.

Future:
- navigate directories
- list directories

'''
class FileClient():
    def __init__(self, server_address):
        self.service = ServiceConsumer('trodes.file.service', server_address=server_address)

    '''
    Opens a file for recording.
    '''
    def open(self, filename):
        # this can be replaced with objects instead of dicts later
        req = {'command': 'open', 'filename': filename}
        self.service.request(req)

    '''
    Starts recording on the open file
    '''
    def start(self):
        # this can be replaced with objects instead of dicts later
        req = {'command': 'start', 'filename': ''}
        self.service.request(req)

    '''
    Pauses recording on the open file
    '''
    def pause(self):
        # this can be replaced with objects instead of dicts later
        req = {'command': 'pause', 'filename': ''}
        self.service.request(req)

    '''
    Closes a file for recording. A file is assumed to already be open.
    '''
    def close(self):
        # this can be replaced with objects instead of dicts later
        req = {'command': 'close', 'filename': ''}
        self.service.request(req)

class FileStatusSubscriber():
    pass


from ..socket import SinkSubscriber, SourcePublisher, SinkPublisher, SourceSubscriber

import threading

'''
Has an inbox... has a sending thread.

Can only be one on a server.
'''
class EventServer:
    def __init__(self, server_address):
        self.thread = EventServer.EventServerThread(server_address)
        self.thread.start()

    class EventServerThread(threading.Thread):
        def __init__(self, server_address):
            super().__init__(daemon=True)
            self.receiver = SinkSubscriber(
                'trodes.event.inbox',
                server_address=server_address)
            self.publisher = SourcePublisher(
                'trodes.event',
                server_address=server_address)
        
        def run(self):
            while True:
                data = self.receiver.receive()
                self.publisher.publish(data)

class EventSender:
    def __init__(self, server_address, event_name):
        self.event_name = event_name

        self.publisher = SinkPublisher(
            'trodes.event.inbox',
            server_address=server_address)

    def publish(self, localTimestamp, systemTimestamp):
        self.publisher.publish({'name': self.event_name, 'localTimestamp': localTimestamp, 'systemTimestamp': systemTimestamp})

class EventSubscription:
    def __init__(self, server_address, event_name):
        self.event_name = event_name

        self.subscriber = SourceSubscriber(
            'trodes.event',
            server_address=server_address)

    def receive(self):
        res = self.subscriber.receive()
        while res['name'] != self.event_name:
            res = self.subscriber.receive()
        return res

class EventSubscriber:
    def __init__(self, server_address, event_name, callback):
        self.thread = EventSubscriber.EventSubscriberThread(server_address, event_name, callback)
        self.thread.start()

    class EventSubscriberThread(threading.Thread):
        def __init__(self, server_address, event_name, callback):
            super().__init__(daemon=True)
            self.subscriber = EventSubscription(server_address, event_name)
            self.callback = callback
        
        def run(self):
            while True:
                res = self.subscriber.receive()
                self.callback(res)



import threading
from ..socket import ServiceConsumer, SourceSubscriber

'''
Use this class to request changes in the acquisition state.

Requires a server_address to connect to the server.
'''
class AcquisitionClient():
    def __init__(self, server_address):
        self.service = ServiceConsumer('trodes.acquisition.service', server_address=server_address)

    def __request(self, command, time):
        # this can be replaced with objects instead of dicts later
        req = {'command': command, 'timestamp': time}
        self.service.request(req)

    def play(self):
        self.__request('play', 0)

    def pause(self):
        self.__request('pause', 0)

    def stop(self):
        self.__request('stop', 0)

    def seek(self, timestamp):
        self.__request('seek', timestamp)
        pass

'''
Requires callback(timestamp).
'''
class AcquisitionSubscriber():
    def __init__(self, server_address, callback):
        self.server_address = server_address
        self.callback = callback

        self.thread = AcquisitionSubscriber.AcquisitionSubscriberThread(self.server_address, callback)
        self.thread.start()
    
    class AcquisitionSubscriberThread(threading.Thread):
        def __init__(self, server_address, callback):
            super().__init__(daemon=True)
            self.subscriber = SourceSubscriber('trodes.acquisition', server_address=server_address)
            self.callback = callback

        def run(self):
            while True:
                res = self.subscriber.receive()

                if res['command'] == 'seek':
                    timestamp = res['timestamp']
                    self.callback(timestamp)



import threading
from ..socket import SourceSubscriber
from .hardware_channel_map import HardwareChannelMap
import struct

'''
Use this class to subscribe to analog sources

`channel_map` is just a dictionary of the channels XML config.
'''
class AnalogSubscription():
    def __init__(self, server_address, channel_map, channel_name):
        self.channel_map = HardwareChannelMap(channel_map)

        device_idx, channel_idx = self.channel_map.find_channel(channel_name)
        id_byte, id_bit, start_byte = self.channel_map.calculate_analog_index(device_idx, channel_idx)

        self.raw_subscriber = RawAnalogSubscription(
            server_address=server_address,
            id_byte=id_byte,
            id_bit=id_bit,
            start_byte=start_byte)

    def receive(self):
        return self.raw_subscriber.receive()

'''
index is the 0-based index of the analog channels 
'''
class RawAnalogSubscription():
    def __init__(self, server_address, id_byte, id_bit, start_byte):
        self.id_byte = id_byte
        self.id_bit = id_bit
        self.start_byte = start_byte

        self.subscriber = SourceSubscriber('source.analog', server_address=server_address)

    def receive(self):
        rec = self.subscriber.receive()
        # merge segments together
        data = bytes().join(rec['analogData'])

        if data[self.id_byte] & (1 << self.id_bit):
            value, = struct.unpack('>h', data[self.start_byte:self.start_byte+2])
            timestamp = rec['localTimestamp']
            return timestamp, value
        else:
            # make this iterative, not recursive
            return self.receive()

'''
Subscriber wraps subscription in a thread and callback

Callback can be used to call a Qt signal
'''
class AnalogSubscriber():
    def __init__(self, server_address, channel_map, channel_name, callback):
        self.thread = AnalogSubscriber.AnalogSubscriberThread(
            server_address, channel_map, channel_name, callback)
        self.thread.start()

    class AnalogSubscriberThread(threading.Thread):
            def __init__(self, server_address, channel_map, channel_name, callback):
                super().__init__(daemon=True)
                self.subscriber = AnalogSubscription(server_address, channel_map, channel_name)
                self.callback = callback

            def run(self):
                while True:
                    res = self.subscriber.receive()
                    self.callback(res)



from .acquisition import *
from .analog import *
from .digital import *
from .data import *
from .events import *
from .file import *
from .hardware_channel_map import *
from .info_client import *
from .trodes_client import *
from .trodes import *



'''
Use this class to query basic information about Trodes.

Requires a network_address to connect to the server.
'''
class TrodesClient():
    def __init__(self, server_address):
        self.server_address = server_address



from .. import socket

from enum import Enum, auto

__all__ = ['CurrentScaling', 'GlobalStimulationSettings', 'StimulationCommand',
           'TrodesHardware', 'TrodesInfoRequester', 'TrodesAnnotationRequester',
           'TrodesAcquisitionRequester', 'TrodesEventSubscriber',
           'TrodesAcquisitionSubscriber', 'TrodesSourceStatusSubscriber']

class CurrentScaling(Enum):
    max10nA = auto()
    max20nA = auto()
    max50nA = auto()
    max100nA = auto()
    max200nA = auto()
    max500nA = auto()
    max1uA = auto()
    max2uA = auto()
    max5uA = auto()
    max10uA = auto()

class GlobalStimulationSettings:
    def setVoltageScale(self, scaleValue):
        self.scaleValue = scaleValue

class StimulationCommand:
    def setBiphasicPulseShape(self, leadingPulseWidth_Samples,
            leadingPulseAmplitude, secondPulseWidth_Samples,
            secondPulseAmplitude, interPhaseDwell_Samples, pulsePeriod_Samples,
            startDelay_Samples):
        self.leadingPulseWidth_Samples = leadingPulseWidth_Samples
        self.leadingPulseAmplitude = leadingPulseAmplitude
        self.secondPulseWidth_Samples = secondPulseWidth_Samples
        self.secondPulseAmplitude = secondPulseAmplitude
        self.interPhaseDwell_Samples = interPhaseDwell_Samples
        self.pulsePeriod_Samples = pulsePeriod_Samples
        self.startDelay_Samples = startDelay_Samples

    def setNumPulsesInTrain(self, numPulsesInTrain):
        self.numPulsesInTrain = numPulsesInTrain

    def setChannels(self, cathodeID, cathodeChannel, anodeID, anodeChannel):
        self.cathodeChannel = cathodeChannel
        self.anodeChannel = anodeChannel
        self.cathodeNtrodeID = cathodeID
        self.anodeNtrodeID = anodeID

    def setGroup(self, group):
        self.group = group

    def setSlot(self, slot):
        self.slot = slot

class TrodesHardware:
    def __init__(self, *, server_address="tcp://127.0.0.1:49152"):
        self.service = socket.ServiceConsumer(
            'trodes.hardware', server_address=server_address)

    def settle_command_triggered(self):
        data = ['tag', 'HRSettle']
        return self.service.request(data)

    def __startstop(self, startstop, slotgroup, number):
        data = [
            'tag',
            'HRStartStopCommand',
            {'startstop': startstop, 'slotgroup': slotgroup, 'number': number}
            ]
        return self.service.request(data)

    def sendStimulationStartSlot(self, slot):
        return self.__startstop('START', 'SLOT', slot)

    def sendStimulationStartGroup(self, group):
        return self.__startstop('START', 'GROUP', group)

    def sendStimulationStopSlot(self, slot):
        return self.__startstop('STOP', 'SLOT', slot)

    def sendStimulationStopGroup(self, group):
        return self.__startstop('STOP', 'GROUP', group)

    def sendStimulationParams(self, params):
        '''
        Takes StimulationCommand params
        '''
        data = [
            'tag',
            'HRSet',
            {
                '_group': params.group,
                'slot': params.slot,
                'cathodeChannel': params.cathodeChannel,
                'anodeChannel': params.anodeChannel,
                'cathodeNtrodeID': params.cathodeNtrodeID,
                'anodeNtrodeID': params.anodeNtrodeID,
                'leadingPulseWidth_Samples': params.leadingPulseWidth_Samples,
                'leadingPulseAmplitude': params.leadingPulseAmplitude,
                'secondPulseWidth_Samples': params.secondPulseWidth_Samples,
                'secondPulseAmplitude': params.secondPulseAmplitude,
                'interPhaseDwell_Samples': params.interPhaseDwell_Samples,
                'pulsePeriod_Samples': params.pulsePeriod_Samples,
                'startDelay_Samples': params.startDelay_Samples,
                'numPulsesInTrain': params.numPulsesInTrain
            }
            ]
        return self.service.request(data)

    def sendClearStimulationParams(self, slot):
        '''
        clear any existing commands in the given slot
        '''
        data = [
            'tag',
            'HRClear',
            { 'number': slot }
            ]
        return self.service.request(data)

    def sendGlobalStimulationSettings(self, settings):
        def getScaleValue(scaleValue):
            if scaleValue == CurrentScaling.max10nA:
                return 'max10nA'
            elif scaleValue == CurrentScaling.max20nA:
                return 'max20nA'
            elif scaleValue == CurrentScaling.max50nA:
                return 'max50nA'
            elif scaleValue == CurrentScaling.max100nA:
                return 'max100nA'
            elif scaleValue == CurrentScaling.max200nA:
                return 'max200nA'
            elif scaleValue == CurrentScaling.max500nA:
                return 'max500nA'
            elif scaleValue == CurrentScaling.max1uA:
                return 'max1uA'
            elif scaleValue == CurrentScaling.max2uA:
                return 'max2uA'
            elif scaleValue == CurrentScaling.max5uA:
                return 'max5uA'
            elif scaleValue == CurrentScaling.max10uA:
                return 'max10uA'
            else:
                raise ValueError("unknown scaleValue enum")

        data = [
            'tag',
            'HRSetGS',
            { 'scaleValue': getScaleValue(settings.scaleValue) }
            ]
        return self.service.request(data)

    def global_stimulation_command(self, resetSequencerCmd,
            killStimulationCmd, clearDSPOffsetRemovalCmd,
            enableStimulation):
        data = [
            'tag',
            'HRSetGC',
            {
                'resetSequencerCmd': resetSequencerCmd,
                'killStimulationCmd': killStimulationCmd,
                'clearDSPOffsetRemovalCmd': clearDSPOffsetRemovalCmd,
                'enableStimulation': enableStimulation,
            }
            ]
        return self.service.request(data)

    def ecu_shortcut_message(self, fn):
        data = [
            'tag',
            'HRSCTrig',
            { 'fn': fn }
            ]
        return self.service.request(data)

class TrodesInfoRequester:
    def __init__(self, *, server_address="tcp://127.0.0.1:49152"):
        self.service = socket.ServiceConsumer(
            'trodes.info', server_address=server_address)

    def __request(self, item):
        data = { 'request': item }
        return self.service.request(data)

    def request_time(self):
        return self.__request('time')[2]['time']

    def request_timerate(self):
        return self.__request('timerate')[2]['timerate']

    def request_config(self):
        return self.__request('config')

class TrodesAnnotationRequester:
    def __init__(self, *, server_address="tcp://127.0.0.1:49152"):
        self.service = socket.ServiceConsumer(
            'trodes.annotation', server_address=server_address)

    def request_annotation(self, timestamp, sender, event):
        data = {
            'timestamp': timestamp,
            'sender': sender,
            'event': event
        }
        return self.service.request(data)

class TrodesAcquisitionRequester:
    def __init__(self, *, server_address="tcp://127.0.0.1:49152"):
        self.service = socket.ServiceConsumer(
            'trodes.acquisition.service', server_address=server_address)

    def __request(self, command, timestamp):
        data = { 'command': command, 'timestamp': timestamp }
        return self.service.request(data)

    def request_play(self):
        return self.__request('play', 0)

    def request_pause(self):
        return self.__request('pause', 0)

    def request_stop(self):
        return self.__request('stop', 0)

    def request_seek(self, timestamp):
        return self.__request('seek', timestamp)

class TrodesEventSubscriber:
    def __init__(self, *, server_address="tcp://127.0.0.1:49152"):
        self.subscriber = socket.SourceSubscriber(
            'trodes.events', server_address=server_address)

    def receive(self, *, noblock=False):
        return self.subscriber.receive(noblock=noblock)

class TrodesAcquisitionSubscriber:
    def __init__(self, *, server_address="tcp://127.0.0.1:49152"):
        self.subscriber = socket.SourceSubscriber(
            'trodes.acquisition', server_address=server_address)

    def receive(self, *, noblock=False):
        return self.subscriber.receive(noblock=noblock)

class TrodesSourceStatusSubscriber:
    def __init__(self, *, server_address="tcp://127.0.0.1:49152"):
        self.subscriber = socket.SourceSubscriber(
            'trodes.source.pub', server_address=server_address)

    def receive(self, *, noblock=False):
        return self.subscriber.receive(noblock=noblock)



import zmq
import msgpack
import time

__all__ = ['SourceSubscriber', 'SourcePublisher', 'SinkPublisher', 'SinkSubscriber', 'ServiceConsumer']

class Socket:
    def __init__(self, type_, *, endpoint=None):
        if type_ not in [zmq.SUB, zmq.PUB, zmq.PUSH, zmq.PULL, zmq.REQ, zmq.REP]:
            raise ValueError('Unknown ZeroMQ socket type. Example: `zmq.SUB`')
        self.context = zmq.Context.instance()
        self.socket = self.context.socket(type_)
        if type_ in [zmq.PUB, zmq.PULL, zmq.REP]:
            if endpoint is None:
                self.socket.bind_to_random_port('tcp://127.0.0.1')
            else:
                self.socket.bind(endpoint)
        else:
            if endpoint == None:
                raise ValueError('An endpoint needs to be specified for this socket type.')
            self.socket.connect(endpoint)
        if type_ == zmq.SUB:
            # By default, we subscribe to all topics
            self.socket.setsockopt(zmq.SUBSCRIBE, b'')
    def get_last_endpoint(self):
        return self.socket.get_string(zmq.LAST_ENDPOINT)
    def send(self, data):
        data_msg = msgpack.packb(data)
        self.socket.send(data_msg)
    def receive(self, *, noblock=False):
        flags = zmq.NOBLOCK if noblock else zmq.NULL
        data_msg = self.socket.recv(flags=flags)
        return msgpack.unpackb(data_msg, raw=False)
    def request(self, request):
        self.send(request)
        return self.receive()

def connection(network_string):
    socket = Socket(zmq.SUB, endpoint=network_string)
    return socket.receive()

def request_endpoint(name, req, try_endpoint):
    conn = connection(try_endpoint)
    socket = Socket(zmq.REQ, endpoint=conn['replyEndpoint'])
    return socket.request(req)

def get_endpoint(name, try_endpoint):
    req = {
        'tag': 'get',
        'name': name,
        'endpoint': ''
        }
    return request_endpoint(name, req, try_endpoint)

def get_endpoint_retry(name, try_endpoint):
    while True:
        endpoint = get_endpoint(name, try_endpoint)
        if (not endpoint == ''):
            break
        print(f'Endpoint `{name}` is not available on the network yet. Retrying in 500ms...')
        time.sleep(0.5)
    return endpoint

def add_endpoint(name, endpoint, server_endpoint):
    req = {
        'tag': 'add',
        'name': name,
        'endpoint': endpoint
        }
    return request_endpoint(name, req, server_endpoint)

class SourceSubscriber:
    def __init__(self, name, *, server_address="tcp://127.0.0.1:49152"):
        endpoint = get_endpoint_retry(name, server_address)
        self.socket = Socket(zmq.SUB, endpoint=endpoint)

    def receive(self, *, noblock=False):
        return self.socket.receive(noblock=noblock)

class SourcePublisher:
    def __init__(self, name, *, endpoint=None,
        server_address="tcp://127.0.0.1:49152"):
        self.socket = Socket(zmq.PUB, endpoint=endpoint)
        add_endpoint(name, self.socket.get_last_endpoint(), server_address)

    def publish(self, data):
        self.socket.send(data)

class SinkPublisher:
    def __init__(self, name, *, server_address="tcp://127.0.0.1:49152"):
        endpoint = get_endpoint_retry(name, server_address)
        self.socket = Socket(zmq.PUSH, endpoint=endpoint)

    def publish(self, data):
        self.socket.send(data)

class SinkSubscriber:
    def __init__(self, name, *, endpoint=None,
        server_address="tcp://127.0.0.1:49152"):
        self.socket = Socket(zmq.PULL, endpoint=endpoint)
        add_endpoint(name, self.socket.get_last_endpoint(), server_address)

    def receive(self, *, noblock=False):
        return self.socket.receive(noblock=noblock)

class ServiceConsumer:
    def __init__(self, name, *,
        server_address="tcp://127.0.0.1:49152"):
        endpoint = get_endpoint_retry(name, server_address)
        self.socket = Socket(zmq.REQ, endpoint=endpoint)

    def request(self, request):
        return self.socket.request(request)





'''
Represents the XML HardwareConfiguration element 

Includes functions for calculating channels.

Stores no state.
'''
class HardwareChannelMap:
    def __init__(self, channel_map):
        # sort devices based on preference
        self.devices = list(channel_map['HardwareConfiguration']['Device'])
        self.devices.sort(key=lambda x: int(x['_packetOrderPreference']))

    def find_channel(self, channel_name):
        # search channels for matching id
        for device_idx, device in enumerate(self.devices):
            for channel_idx, channel in enumerate(device['Channel']):
                if channel['_id'] == channel_name:
                    return device_idx, channel_idx
        raise KeyError('Could not find channel name in channel map.')
    
    def calculate_digital_index(self, device_idx, channel_idx):
        # offset by the devices coming before
        sync_byte_offset = 1
        device_byte_offset = sum(map(lambda x: int(x['_numBytes']), self.devices[:device_idx]))

        # offset by byte position in device
        channel = self.devices[device_idx]['Channel'][channel_idx]
        channel_byte_offset = int(channel['_startByte'])
        channel_bit = int(channel['_bit'])

        return (sync_byte_offset + device_byte_offset + channel_byte_offset) * 8 + channel_bit
    
    '''
    Assumes 1) fetching interleaved analog data 2) analog channels from previous devices need offset
    '''
    def calculate_analog_index(self, device_idx, channel_idx):
        prev_devices = self.devices[:device_idx]
        prev_devices_channels = [chan for dev in prev_devices for chan in dev['Channel']]
        prev_analog_channels = list(filter(lambda x: self.is_analog(x), prev_devices_channels))

        # calculate the uninterleaved from previous devices
        # each analog channel is 2 bytes
        uninterleaved_offset = len(prev_analog_channels) * 2

        channel = self.devices[device_idx]['Channel'][channel_idx]
        channel_id_byte = uninterleaved_offset + int(channel['_interleavedDataIDByte'])
        channel_id_bit = int(channel['_interleavedDataIDBit'])
        channel_start_byte = uninterleaved_offset + int(channel['_startByte'])

        return channel_id_byte, channel_id_bit, channel_start_byte

    def is_analog(self, chan):
        return chan['_dataType'] == 'analog'


import threading
from ..socket import ServiceConsumer

'''
Allows you to enable and disable certain data streams over the network.
'''
class DataClient():
    def __init__(self, server_address):
        self.service = ServiceConsumer('trodes.data.service', server_address=server_address)

    def enable(self, streamname):
        req = {'command': 'enable', 'streamname': streamname}
        self.service.request(req)

    def disable(self, streamname):
        req = {'command': 'disable', 'streamname': streamname}
        self.service.request(req)



from ..socket import SourceSubscriber
from .hardware_channel_map import HardwareChannelMap

import threading

'''
Use this class to subscribe to analog sources

Requires input of a channel map object. The channel map is just
a JSON-like dictionary of the XML HardwareConfiguration node
in the `.trodesconfig` file.

Requires a server_address to connect to the server.

It can be used like this:

    subscriber = trodes.DigitalClient(
        server_address=self.network_address,
        channel_map=config.channel_map,
        channel_name='ECU_Din8')

'''
class DigitalClient():
    def __init__(self, server_address, channel_map, channel_name):
        self.channel_map = HardwareChannelMap(channel_map)

        device_idx, channel_idx = self.channel_map.find_channel(channel_name)
        index =  self.channel_map.calculate_digital_index(device_idx, channel_idx)

        self.raw_client = RawDigitalClient(server_address=server_address, index=index)

    def receive(self):
        return self.raw_client.receive()

class RawDigitalClient():
    def __init__(self, server_address, index):
        self.index = index
        self.byte_id = index // 8
        self.bit_id = index % 8

        self.subscriber = SourceSubscriber('source.digital', server_address=server_address)

    def receive(self):
        rec = self.subscriber.receive()
        timestamp = rec['localTimestamp']
        data = rec['digitalData'][0]
        bit = (data[self.byte_id] >> self.bit_id) & 1
        return timestamp, bit

'''
Subscriber wraps subscription in a thread and callback

Callback can be used to call a Qt signal
'''
class DigitalSubscriber():
    def __init__(self, server_address, channel_map, channel_name, callback):
        self.thread = DigitalSubscriber.DigitalSubscriberThread(
            server_address, channel_map, channel_name, callback)
        self.thread.start()

    class DigitalSubscriberThread(threading.Thread):
            def __init__(self, server_address, channel_map, channel_name, callback):
                super().__init__(daemon=True)
                self.subscriber = DigitalClient(server_address, channel_map, channel_name)
                self.callback = callback

            def run(self):
                while True:
                    res = self.subscriber.receive()
                    self.callback(res)


from setuptools import setup

setup(
    name="trodesnetwork",
    version="0.0.11",
    description="A library to connect to Trodes over a network",
    packages=["trodesnetwork", "trodesnetwork.trodes"],
    install_requires=[
        'pyzmq >=18.0.0,<20.0.0',
        'msgpack'
    ]
)

